"""Unit tests for memorydb-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


def create_subnet_group(client, region_name):
    """Return valid Subnet group."""
    ec2 = boto3.resource("ec2", region_name=region_name)
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet1 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/24")
    subnet2 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.1.0/24")
    subnet_group = client.create_subnet_group(
        SubnetGroupName="my_subnet_group",
        Description="This is my subnet group",
        SubnetIds=[subnet1.id, subnet2.id],
        Tags=[
            {"Key": "foo", "Value": "bar"},
        ],
    )
    return subnet_group


@mock_aws
def test_create_cluster():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    resp = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    cluster = resp["Cluster"]
    assert "Name" in cluster
    assert "Status" in cluster
    assert "NumberOfShards" in cluster


@mock_aws
def test_create_duplicate_cluster_fails():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    client.create_cluster(
        ClusterName="foo-bar",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    with pytest.raises(ClientError) as ex:
        client.create_cluster(
            ClusterName="foo-bar", NodeType="db.t4g.small", ACLName="open-access"
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterAlreadyExistsFault"


@mock_aws
def test_create_subnet_group():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    subnet_group = create_subnet_group(client, "ap-southeast-1")
    sg = subnet_group["SubnetGroup"]
    assert "Name" in sg
    assert "Description" in sg
    assert "VpcId" in sg
    assert "Subnets" in sg
    assert "ARN" in sg


@mock_aws
def test_create_cluster_with_subnet_group():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    subnet_group = create_subnet_group(client, "ap-southeast-1")
    resp = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        SubnetGroupName=subnet_group["SubnetGroup"]["Name"],
        ACLName="open-access",
    )
    subnet_group = resp["Cluster"]["SubnetGroupName"] == "my_subnet_group"


@mock_aws
def test_create_duplicate_subnet_group_fails():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    create_subnet_group(client, "ap-southeast-1")
    with pytest.raises(ClientError) as ex:
        create_subnet_group(client, "ap-southeast-1")
    err = ex.value.response["Error"]
    assert err["Code"] == "SubnetGroupAlreadyExistsFault"


@mock_aws
def test_create_invalid_subnet_group_fails():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    with pytest.raises(ClientError) as ex:
        client.create_subnet_group(SubnetGroupName="foo-bar", SubnetIds=["foo", "bar"])
    err = ex.value.response["Error"]
    assert err["Code"] == "InvalidSubnetError"


@mock_aws
def test_create_snapshot():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    subnet_group = create_subnet_group(client, "ap-southeast-1")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        Description="Test memorydb cluster",
        NodeType="db.t4g.small",
        SubnetGroupName=subnet_group["SubnetGroup"]["Name"],
        ACLName="open-access",
    )
    resp = client.create_snapshot(
        ClusterName=cluster["Cluster"]["Name"],
        SnapshotName="my-snapshot-1",
        KmsKeyId=f"arn:aws:kms:ap-southeast-1:{ACCOUNT_ID}:key/51d81fab-b138-4bd2-8a09-07fd6d37224d",
        Tags=[
            {"Key": "foo", "Value": "bar"},
        ],
    )
    snapshot = resp["Snapshot"]
    assert "Name" in snapshot
    assert "Status" in snapshot
    assert "Source" in snapshot
    assert "KmsKeyId" in snapshot
    assert "ARN" in snapshot
    assert "ClusterConfiguration" in snapshot
    assert "DataTiering" in snapshot


@mock_aws
def test_create_snapshot_with_non_existing_cluster_fails():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    with pytest.raises(ClientError) as ex:
        client.create_snapshot(ClusterName="foobar", SnapshotName="my-snapshot-1")
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterNotFoundFault"


@mock_aws
def test_create_duplicate_snapshot_fails():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    client.create_snapshot(
        ClusterName=cluster["Cluster"]["Name"], SnapshotName="my-snapshot-1"
    )
    with pytest.raises(ClientError) as ex:
        client.create_snapshot(
            ClusterName=cluster["Cluster"]["Name"], SnapshotName="my-snapshot-1"
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotAlreadyExistsFault"


@mock_aws
def test_describe_clusters():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
    resp = client.describe_clusters()
    assert "Clusters" in resp
    assert len(resp["Clusters"]) == 2
    assert "Shards" not in resp["Clusters"][0]


@mock_aws
def test_describe_clusters_with_shard_details():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
    resp = client.describe_clusters(
        ClusterName="test-memory-db-1",
        ShowShardDetails=True,
    )
    assert resp["Clusters"][0]["Name"] == "test-memory-db-1"
    assert len(resp["Clusters"]) == 1
    assert "Shards" in resp["Clusters"][0]


@mock_aws
def test_describe_clusters_with_cluster_name():
    client = boto3.client("memorydb", region_name="ap-southeast-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
    resp = client.describe_clusters(
        ClusterName="test-memory-db-1",
    )
    assert resp["Clusters"][0]["Name"] == "test-memory-db-1"
    assert len(resp["Clusters"]) == 1


@mock_aws
def test_describe_snapshots():
    client = boto3.client("memorydb", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
        client.create_snapshot(
            ClusterName=f"test-memory-db-{i}", SnapshotName=f"my-snapshot-{i}"
        )
    resp = client.describe_snapshots()
    assert "Snapshots" in resp
    assert len(resp["Snapshots"]) == 2
    assert resp["Snapshots"][0]["Name"] == "my-snapshot-1"


@mock_aws
def test_describe_snapshots_with_cluster_name():
    client = boto3.client("memorydb", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
        client.create_snapshot(
            ClusterName=f"test-memory-db-{i}", SnapshotName=f"my-snapshot-{i}"
        )
    resp = client.describe_snapshots(ClusterName="test-memory-db-2")
    assert len(resp["Snapshots"]) == 1
    assert resp["Snapshots"][0]["ClusterConfiguration"]["Name"] == "test-memory-db-2"
    assert "Shards" not in resp["Snapshots"][0]["ClusterConfiguration"]


@mock_aws
def test_describe_snapshots_with_shard_details():
    client = boto3.client("memorydb", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
        client.create_snapshot(
            ClusterName=f"test-memory-db-{i}", SnapshotName=f"my-snapshot-{i}"
        )
    resp = client.describe_snapshots(ClusterName="test-memory-db-2", ShowDetail=True)
    assert len(resp["Snapshots"]) == 1
    assert resp["Snapshots"][0]["ClusterConfiguration"]["Name"] == "test-memory-db-2"
    assert "Shards" in resp["Snapshots"][0]["ClusterConfiguration"]


@mock_aws
def test_describe_snapshots_with_snapshot_name():
    client = boto3.client("memorydb", region_name="eu-west-1")
    for i in range(1, 3):
        client.create_cluster(
            ClusterName=f"test-memory-db-{i}",
            NodeType="db.t4g.small",
            ACLName="open-access",
        )
        client.create_snapshot(
            ClusterName=f"test-memory-db-{i}", SnapshotName=f"my-snapshot-{i}"
        )
    resp = client.describe_snapshots(
        SnapshotName="my-snapshot-1",
    )
    assert len(resp["Snapshots"]) == 1
    assert resp["Snapshots"][0]["Name"] == "my-snapshot-1"


@mock_aws
def test_describe_snapshots_with_snapshot_and_cluster():
    client = boto3.client("memorydb", region_name="eu-west-1")

    client.create_cluster(
        ClusterName="test-memory-db", NodeType="db.t4g.small", ACLName="open-access"
    )
    for i in range(1, 3):
        client.create_snapshot(
            ClusterName="test-memory-db", SnapshotName=f"my-snapshot-{i}"
        )
    resp = client.describe_snapshots(
        ClusterName="test-memory-db",
        SnapshotName="my-snapshot-1",
    )
    assert len(resp["Snapshots"]) == 1
    assert resp["Snapshots"][0]["Name"] == "my-snapshot-1"


@mock_aws
def test_describe_snapshots_with_invalid_cluster():
    client = boto3.client("memorydb", region_name="eu-west-1")

    resp = client.describe_snapshots(
        ClusterName="foobar",
    )
    assert len(resp["Snapshots"]) == 0


@mock_aws
def test_describe_snapshots_invalid_snapshot_fails():
    client = boto3.client("memorydb", region_name="eu-west-1")

    with pytest.raises(ClientError) as ex:
        client.describe_snapshots(SnapshotName="foobar")
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotNotFoundFault"


@mock_aws
def test_describe_snapshots_with_cluster_and_invalid_snapshot_fails():
    client = boto3.client("memorydb", region_name="eu-west-1")

    client.create_cluster(
        ClusterName="test-memory-db", NodeType="db.t4g.small", ACLName="open-access"
    )
    client.create_snapshot(ClusterName="test-memory-db", SnapshotName="my-snapshot")

    with pytest.raises(ClientError) as ex:
        client.describe_snapshots(ClusterName="test-memory-db", SnapshotName="foobar")
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotNotFoundFault"


@mock_aws
def test_describe_subnet_groups():
    client = boto3.client("memorydb", region_name="eu-west-1")
    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet1 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/24")
    subnet2 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.1.0/24")
    for i in range(1, 3):
        client.create_subnet_group(
            SubnetGroupName=f"my_subnet_group-{i}",
            Description="This is my subnet group",
            SubnetIds=[subnet1.id, subnet2.id],
        )
    resp = client.describe_subnet_groups()
    assert "SubnetGroups" in resp
    assert len(resp["SubnetGroups"]) == 3  # Including default subnet group


@mock_aws
def test_describe_subnet_groups_with_subnet_group_name():
    client = boto3.client("memorydb", region_name="eu-west-1")
    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="10.0.0.0/16")
    subnet1 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.0.0/24")
    subnet2 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="10.0.1.0/24")
    for i in range(1, 3):
        client.create_subnet_group(
            SubnetGroupName=f"my_subnet_group-{i}",
            Description="This is my subnet group",
            SubnetIds=[subnet1.id, subnet2.id],
        )
    resp = client.describe_subnet_groups(SubnetGroupName="my_subnet_group-1")
    assert len(resp["SubnetGroups"]) == 1
    assert resp["SubnetGroups"][0]["Name"] == "my_subnet_group-1"


@mock_aws
def test_describe_subnet_groups_invalid_subnetgroupname_fails():
    client = boto3.client("memorydb", region_name="eu-west-1")
    with pytest.raises(ClientError) as ex:
        client.describe_subnet_groups(SubnetGroupName="foobar")
    err = ex.value.response["Error"]
    assert err["Code"] == "SubnetGroupNotFoundFault"


@mock_aws
def test_list_tags():
    client = boto3.client("memorydb", region_name="us-east-2")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
        Tags=[
            {"Key": "foo", "Value": "bar"},
        ],
    )
    resp = client.list_tags(ResourceArn=cluster["Cluster"]["ARN"])
    assert "TagList" in resp
    assert len(resp["TagList"]) == 1
    assert "foo" in resp["TagList"][0]["Key"]
    assert "bar" in resp["TagList"][0]["Value"]


@mock_aws
def test_list_tags_snapshot():
    client = boto3.client("memorydb", region_name="us-east-2")
    client.create_cluster(
        ClusterName="test-memory-db", NodeType="db.t4g.small", ACLName="open-access"
    )
    snapshot = client.create_snapshot(
        ClusterName="test-memory-db",
        SnapshotName="my-snapshot",
        Tags=[
            {"Key": "foo1", "Value": "bar1"},
        ],
    )
    resp = client.list_tags(ResourceArn=snapshot["Snapshot"]["ARN"])
    assert len(resp["TagList"]) == 1


@mock_aws
def test_list_tags_subnetgroups():
    client = boto3.client("memorydb", region_name="us-east-2")
    subnet_group = create_subnet_group(client, "us-east-2")
    sg = subnet_group["SubnetGroup"]
    resp = client.list_tags(ResourceArn=sg["ARN"])
    assert len(resp["TagList"]) == 1


@mock_aws
def test_list_tags_invalid_cluster_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.list_tags(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:cluster/foobar",
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterNotFoundFault"


@mock_aws
def test_list_tags_invalid_snapshot_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.list_tags(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:snapshot/foobar",
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotNotFoundFault"


@mock_aws
def test_list_tags_invalid_subnetgroup_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.list_tags(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:subnetgroup/foobar",
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SubnetGroupNotFoundFault"


@mock_aws
def test_tag_resource():
    client = boto3.client("memorydb", region_name="us-east-2")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
        Tags=[
            {"Key": "key1", "Value": "value1"},
        ],
    )
    resp = client.tag_resource(
        ResourceArn=cluster["Cluster"]["ARN"],
        Tags=[
            {"Key": "key2", "Value": "value2"},
        ],
    )
    assert "TagList" in resp
    assert len(resp["TagList"]) == 2
    assert "key2" in resp["TagList"][1]["Key"]
    assert "value2" in resp["TagList"][1]["Value"]


@mock_aws
def test_tag_resource_invalid_cluster_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.tag_resource(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:cluster/foobar",
            Tags=[{"Key": "key2", "Value": "value2"}],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterNotFoundFault"


@mock_aws
def test_tag_resource_invalid_snapshot_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.tag_resource(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:snapshot/foobar",
            Tags=[{"Key": "key2", "Value": "value2"}],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotNotFoundFault"


@mock_aws
def test_tag_resource_invalid_subnetgroup_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.tag_resource(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:subnetgroup/foobar",
            Tags=[{"Key": "key2", "Value": "value2"}],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SubnetGroupNotFoundFault"


@mock_aws
def test_untag_resource():
    client = boto3.client("memorydb", region_name="us-east-2")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
        Tags=[{"Key": "key1", "Value": "value1"}, {"Key": "key2", "Value": "value2"}],
    )
    resp = client.untag_resource(
        ResourceArn=cluster["Cluster"]["ARN"],
        TagKeys=[
            "key1",
        ],
    )
    assert "TagList" in resp
    assert len(resp["TagList"]) == 1
    assert "key2" in resp["TagList"][0]["Key"]
    assert "value2" in resp["TagList"][0]["Value"]


@mock_aws
def test_untag_resource_invalid_cluster_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.untag_resource(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:cluster/foobar",
            TagKeys=["key1"],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterNotFoundFault"


@mock_aws
def test_untag_resource_invalid_snapshot_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.untag_resource(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:snapshot/foobar",
            TagKeys=["key1"],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotNotFoundFault"


@mock_aws
def test_untag_resource_invalid_subnetgroup_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.untag_resource(
            ResourceArn=f"arn:aws:memorydb:us-east-1:{ACCOUNT_ID}:subnetgroup/foobar",
            TagKeys=["key1"],
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "SubnetGroupNotFoundFault"


@mock_aws
def test_untag_resource_invalid_keys_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
        Tags=[{"Key": "key1", "Value": "value1"}, {"Key": "key2", "Value": "value2"}],
    )
    with pytest.raises(ClientError) as ex:
        client.untag_resource(
            ResourceArn=cluster["Cluster"]["ARN"], TagKeys=["key3", "key4"]
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "TagNotFoundFault"


@mock_aws
def test_update_cluster_replica_count():
    client = boto3.client("memorydb", region_name="eu-west-1")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    desc_before_update = client.describe_clusters(ShowShardDetails=True)
    assert desc_before_update["Clusters"][0]["Shards"][0]["NumberOfNodes"] == 2
    client.update_cluster(
        ClusterName=cluster["Cluster"]["Name"],
        Description="Good cluster",
        MaintenanceWindow="thu:23:00-thu:01:30",
        ReplicaConfiguration={"ReplicaCount": 2},
    )
    desc_after_update = client.describe_clusters(ShowShardDetails=True)
    cluster_after_update = desc_after_update["Clusters"][0]
    assert cluster_after_update["Description"] == "Good cluster"
    assert cluster_after_update["MaintenanceWindow"] == "thu:23:00-thu:01:30"
    assert cluster_after_update["Shards"][0]["NumberOfNodes"] == 3


@mock_aws
def test_update_cluster_shards():
    client = boto3.client("memorydb", region_name="eu-west-1")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    desc_before_update = client.describe_clusters(ShowShardDetails=True)
    assert desc_before_update["Clusters"][0]["NumberOfShards"] == 1
    client.update_cluster(
        ClusterName=cluster["Cluster"]["Name"],
        ShardConfiguration={"ShardCount": 2},
    )
    desc_after_update = client.describe_clusters(ShowShardDetails=True)
    assert desc_after_update["Clusters"][0]["NumberOfShards"] == 2


@mock_aws
def test_update_invalid_cluster_fails():
    client = boto3.client("memorydb", region_name="eu-west-1")
    with pytest.raises(ClientError) as ex:
        client.update_cluster(
            ClusterName="foobar",
            Description="Good cluster",
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterNotFoundFault"


@mock_aws
def test_delete_cluster():
    client = boto3.client("memorydb", region_name="eu-west-1")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    desc_resp_before = client.describe_clusters()
    assert len(desc_resp_before["Clusters"]) == 1
    resp = client.delete_cluster(
        ClusterName=cluster["Cluster"]["Name"],
    )
    assert resp["Cluster"]["Name"] == cluster["Cluster"]["Name"]
    desc_resp_after = client.describe_clusters()
    assert len(desc_resp_after["Clusters"]) == 0


@mock_aws
def test_delete_cluster_with_snapshot():
    client = boto3.client("memorydb", region_name="eu-west-1")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    desc_resp_before = client.describe_snapshots()
    assert len(desc_resp_before["Snapshots"]) == 0
    resp = client.delete_cluster(
        ClusterName=cluster["Cluster"]["Name"],
        FinalSnapshotName="test-memory-db-snapshot",
    )
    assert resp["Cluster"]["Name"] == cluster["Cluster"]["Name"]
    desc_resp_after = client.describe_snapshots()
    assert len(desc_resp_after["Snapshots"]) == 1
    assert desc_resp_after["Snapshots"][0]["Name"] == "test-memory-db-snapshot"


@mock_aws
def test_delete_invalid_cluster_fails():
    client = boto3.client("memorydb", region_name="eu-west-1")
    with pytest.raises(ClientError) as ex:
        client.delete_cluster(
            ClusterName="foobar",
        )
    err = ex.value.response["Error"]
    assert err["Code"] == "ClusterNotFoundFault"


@mock_aws
def test_delete_snapshot():
    client = boto3.client("memorydb", region_name="us-east-2")
    cluster = client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        ACLName="open-access",
    )
    snapshot = client.create_snapshot(
        ClusterName=cluster["Cluster"]["Name"],
        SnapshotName="my-snapshot-1",
    )
    desc_resp_before = client.describe_snapshots()
    assert len(desc_resp_before["Snapshots"]) == 1
    resp = client.delete_snapshot(SnapshotName=snapshot["Snapshot"]["Name"])
    assert "Snapshot" in resp
    desc_resp_after = client.describe_snapshots()
    assert len(desc_resp_after["Snapshots"]) == 0


@mock_aws
def test_delete_invalid_snapshot_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    with pytest.raises(ClientError) as ex:
        client.delete_snapshot(SnapshotName="foobar")
    err = ex.value.response["Error"]
    assert err["Code"] == "SnapshotNotFoundFault"


@mock_aws
def test_delete_subnet_group():
    client = boto3.client("memorydb", region_name="us-east-2")
    subnet_group = create_subnet_group(client, "us-east-2")
    sg = subnet_group["SubnetGroup"]
    response = client.describe_subnet_groups()
    assert len(response["SubnetGroups"]) == 2
    resp = client.delete_subnet_group(SubnetGroupName=sg["Name"])
    assert "SubnetGroup" in resp
    response = client.describe_subnet_groups()
    assert len(response["SubnetGroups"]) == 1  # default subnet group


@mock_aws
def test_delete_subnet_group_default_fails():
    client = boto3.client("memorydb", region_name="us-east-2")

    with pytest.raises(ClientError) as ex:
        client.delete_subnet_group(SubnetGroupName="default")
    err = ex.value.response["Error"]
    assert err["Code"] == "InvalidParameterValueException"


@mock_aws
def test_delete_subnet_group_in_use_fails():
    client = boto3.client("memorydb", region_name="us-east-2")
    subnet_group = create_subnet_group(client, "us-east-2")
    client.create_cluster(
        ClusterName="test-memory-db",
        NodeType="db.t4g.small",
        SubnetGroupName=subnet_group["SubnetGroup"]["Name"],
        ACLName="open-access",
    )
    with pytest.raises(ClientError) as ex:
        client.delete_subnet_group(SubnetGroupName=subnet_group["SubnetGroup"]["Name"])
    err = ex.value.response["Error"]
    assert err["Code"] == "SubnetGroupInUseFault"



from __future__ import annotations

import json
import re
from typing import Any, Callable, Dict, List, Optional, Set

from moto.stepfunctions.parser.asl.utils.encoding import to_json_str

JSONataExpression = str
VariableReference = str
VariableDeclarations = str

_PATTERN_VARIABLE_REFERENCE: re.Pattern = re.compile(
    r"\$\$|\$[a-zA-Z0-9_$]+(?:\.[a-zA-Z0-9_][a-zA-Z0-9_$]*)*|\$"
)
_ILLEGAL_VARIABLE_REFERENCES: Set[str] = {"$", "$$"}
_VARIABLE_REFERENCE_ASSIGNMENT_OPERATOR: str = ":="
_VARIABLE_REFERENCE_ASSIGNMENT_STOP_SYMBOL: str = ";"
_EXPRESSION_OPEN_SYMBOL: str = "("
_EXPRESSION_CLOSE_SYMBOL: str = ")"


class JSONataException(Exception):
    error: str
    details: Optional[str]

    def __init__(self, error: str, details: Optional[str]):
        self.error = error
        self.details = details


class _JSONataJVMBridge:
    _java_OBJECT_MAPPER: "com.fasterxml.jackson.databind.ObjectMapper"  # noqa
    _java_JSONATA: "com.dashjoin.jsonata.Jsonata.jsonata"  # noqa

    def __init__(self):
        pass

    @staticmethod
    def get() -> _JSONataJVMBridge:
        return _JSONataJVMBridge()

    def eval_jsonata(self, jsonata_expression: JSONataExpression) -> Any:
        try:
            # Evaluate the JSONata expression with the JVM.
            # TODO: Investigate whether it is worth moving this chain of statements (java_*) to a
            #  Java program to reduce i/o between the JVM and this runtime.
            java_expression = self._java_JSONATA(jsonata_expression)
            java_output = java_expression.evaluate(None)
            java_output_string = self._java_OBJECT_MAPPER.writeValueAsString(
                java_output
            )

            # Compute a Python json object from the java string, this is to:
            #  1. Ensure we fully end interactions with the JVM about this value here;
            #  2. The output object may undergo under operations that are not compatible
            #     with jpype objects (such as json.dumps, equality, instanceof, etc.).
            result_str: str = str(java_output_string)
            result_json = json.loads(result_str)

            return result_json
        except Exception as ex:
            raise JSONataException("UNKNOWN", str(ex))


# Final reference to the java evaluation function.
_eval_jsonata: Callable[[JSONataExpression], Any] = None


def eval_jsonata_expression(jsonata_expression: JSONataExpression) -> Any:
    return _eval_jsonata(jsonata_expression)


class IllegalJSONataVariableReference(ValueError):
    variable_reference: VariableReference

    def __init__(self, variable_reference: VariableReference):
        self.variable_reference = variable_reference


def extract_jsonata_variable_references(
    jsonata_expression: JSONataExpression,
) -> set[VariableReference]:
    if not jsonata_expression:
        return set()
    variable_references: List[VariableReference] = _PATTERN_VARIABLE_REFERENCE.findall(
        jsonata_expression
    )
    for variable_reference in variable_references:
        if variable_reference in _ILLEGAL_VARIABLE_REFERENCES:
            raise IllegalJSONataVariableReference(variable_reference=variable_reference)
    return set(variable_references)


def encode_jsonata_variable_declarations(
    bindings: Dict[VariableReference, Any],
) -> VariableDeclarations:
    declarations_parts: List[str] = list()
    for variable_reference, value in bindings.items():
        if isinstance(value, str):
            value_str_lit = f'"{value}"'
        else:
            value_str_lit = to_json_str(value, separators=(",", ":"))
        declarations_parts.extend(
            [
                variable_reference,
                _VARIABLE_REFERENCE_ASSIGNMENT_OPERATOR,
                value_str_lit,
                _VARIABLE_REFERENCE_ASSIGNMENT_STOP_SYMBOL,
            ]
        )
    return "".join(declarations_parts)


def compose_jsonata_expression(
    final_jsonata_expression: JSONataExpression,
    variable_declarations_list: List[VariableDeclarations],
) -> JSONataExpression:
    # TODO: should be expanded to pack the intrinsic functions too.
    variable_declarations = "".join(variable_declarations_list)
    expression = "".join(
        [
            _EXPRESSION_OPEN_SYMBOL,
            variable_declarations,
            final_jsonata_expression,
            _EXPRESSION_CLOSE_SYMBOL,
        ]
    )
    return expression



"""Unit tests for sagemaker-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_cluster():
    client = boto3.client("sagemaker", region_name="us-east-1")
    resp = client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    assert (
        resp["ClusterArn"]
        == "arn:aws:sagemaker:us-east-1:123456789012:cluster/testcluster"
    )


@mock_aws
def test_describe_cluster():
    client = boto3.client("sagemaker", region_name="us-east-2")
    arn = client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    resp = client.describe_cluster(ClusterName=arn["ClusterArn"])
    assert resp["ClusterArn"] == arn["ClusterArn"]
    assert resp["ClusterName"] == "testcluster"
    assert resp["ClusterStatus"] == "InService"
    assert resp["InstanceGroups"] == [
        {
            "CurrentCount": 10,
            "TargetCount": 10,
            "InstanceGroupName": "testgroup",
            "InstanceType": "ml.p4d.24xlarge",
            "LifeCycleConfig": {
                "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                "OnCreate": "filename",
            },
            "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
            "ThreadsPerCore": 2,
        },
        {
            "CurrentCount": 15,
            "TargetCount": 15,
            "InstanceGroupName": "testgroup2",
            "InstanceType": "ml.g5.8xlarge",
            "LifeCycleConfig": {
                "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                "OnCreate": "filename",
            },
            "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
            "ThreadsPerCore": 1,
        },
    ]
    assert resp["VpcConfig"] == {
        "SecurityGroupIds": [
            "sg-12345678901234567",
        ],
        "Subnets": [
            "subnet-12345678901234567",
        ],
    }


@mock_aws
def test_delete_cluster():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    arn = client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    resp = client.delete_cluster(ClusterName=arn["ClusterArn"])
    assert resp["ClusterArn"] == arn["ClusterArn"]
    with pytest.raises(ClientError) as e:
        resp = client.describe_cluster(ClusterName=arn["ClusterArn"])
    assert (
        str(e.value)
        == "An error occurred (ValidationException) when calling the DescribeCluster operation: Could not find cluster 'testcluster'."
    )


@mock_aws
def test_describe_cluster_node():
    client = boto3.client("sagemaker", region_name="eu-west-1")
    client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    resp = client.describe_cluster_node(
        ClusterName="testcluster", NodeId="testgroup2-4"
    )
    assert resp["NodeDetails"]["InstanceGroupName"] == "testgroup2"
    assert resp["NodeDetails"]["InstanceId"] == "testgroup2-4"
    assert resp["NodeDetails"]["InstanceStatus"] == {
        "Status": "Running",
        "Message": "message",
    }
    assert resp["NodeDetails"]["InstanceType"] == "ml.g5.8xlarge"
    assert resp["NodeDetails"]["LifeCycleConfig"] == {
        "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
        "OnCreate": "filename",
    }
    assert resp["NodeDetails"]["ThreadsPerCore"] == 1


@mock_aws
def test_list_clusters():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    arn = client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    arn2 = client.create_cluster(
        ClusterName="testcluster2",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup44",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup22",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    resp = client.list_clusters()
    assert resp["ClusterSummaries"][0]["ClusterArn"] == arn["ClusterArn"]
    assert resp["ClusterSummaries"][1]["ClusterArn"] == arn2["ClusterArn"]
    assert resp["ClusterSummaries"][0]["ClusterName"] == "testcluster"
    assert resp["ClusterSummaries"][1]["ClusterName"] == "testcluster2"
    assert resp["ClusterSummaries"][0]["ClusterStatus"] == "InService"
    assert resp["ClusterSummaries"][1]["ClusterStatus"] == "InService"


@mock_aws
def test_list_clusters_filters_sorting():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    arn2 = client.create_cluster(
        ClusterName="testcluster2",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup44",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup23",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )

    arn3 = client.create_cluster(
        ClusterName="testcluster22",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup3",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup25",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )

    resp = client.list_clusters(NameContains="2", SortBy="Name", SortOrder="Ascending")
    assert resp["ClusterSummaries"][0]["ClusterArn"] == arn2["ClusterArn"]
    assert resp["ClusterSummaries"][1]["ClusterArn"] == arn3["ClusterArn"]
    assert resp["ClusterSummaries"][0]["ClusterName"] == "testcluster2"
    assert resp["ClusterSummaries"][1]["ClusterName"] == "testcluster22"
    assert resp["ClusterSummaries"][0]["ClusterStatus"] == "InService"
    assert resp["ClusterSummaries"][1]["ClusterStatus"] == "InService"


@mock_aws
def test_list_cluster_nodes():
    client = boto3.client("sagemaker", region_name="ap-southeast-1")
    client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    resp = client.list_cluster_nodes(
        ClusterName="testcluster",
    )
    assert len(resp["ClusterNodeSummaries"]) == 25
    resp = client.list_cluster_nodes(
        ClusterName="testcluster", InstanceGroupNameContains="testgroup2"
    )
    assert len(resp["ClusterNodeSummaries"]) == 15
    assert resp["ClusterNodeSummaries"][0]["InstanceGroupName"] == "testgroup2"
    assert resp["ClusterNodeSummaries"][0]["InstanceId"] == "testgroup2-0"
    assert resp["ClusterNodeSummaries"][0]["InstanceType"] == "ml.g5.8xlarge"
    assert resp["ClusterNodeSummaries"][0]["InstanceStatus"] == {
        "Status": "Running",
        "Message": "message",
    }


# tagging test
@mock_aws
def test_tag_cluster():
    client = boto3.client("sagemaker", region_name="us-east-1")
    arn = client.create_cluster(
        ClusterName="testcluster5",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    resp = client.add_tags(
        ResourceArn=arn["ClusterArn"],
        Tags=[
            {"Key": "testkey2", "Value": "testvalue2"},
        ],
    )
    resp = client.list_tags(ResourceArn=arn["ClusterArn"])
    assert resp["Tags"] == [
        {"Key": "testkey", "Value": "testvalue"},
        {"Key": "testkey2", "Value": "testvalue2"},
    ]


@mock_aws
def test_create_cluster_duplicate_name():
    client = boto3.client("sagemaker", region_name="us-east-1")
    client.create_cluster(
        ClusterName="testcluster",
        InstanceGroups=[
            {
                "InstanceCount": 10,
                "InstanceGroupName": "testgroup",
                "InstanceType": "ml.p4d.24xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 2,
            },
            {
                "InstanceCount": 15,
                "InstanceGroupName": "testgroup2",
                "InstanceType": "ml.g5.8xlarge",
                "LifeCycleConfig": {
                    "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                    "OnCreate": "filename",
                },
                "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                "ThreadsPerCore": 1,
            },
        ],
        VpcConfig={
            "SecurityGroupIds": [
                "sg-12345678901234567",
            ],
            "Subnets": [
                "subnet-12345678901234567",
            ],
        },
        Tags=[
            {"Key": "testkey", "Value": "testvalue"},
        ],
    )
    with pytest.raises(ClientError) as e:
        client.create_cluster(
            ClusterName="testcluster",
            InstanceGroups=[
                {
                    "InstanceCount": 10,
                    "InstanceGroupName": "testgroup",
                    "InstanceType": "ml.p4d.24xlarge",
                    "LifeCycleConfig": {
                        "SourceS3Uri": "s3://sagemaker-lifecycleconfig",
                        "OnCreate": "filename",
                    },
                    "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                    "ThreadsPerCore": 2,
                },
                {
                    "InstanceCount": 15,
                    "InstanceGroupName": "testgroup2",
                    "InstanceType": "ml.g5.8xlarge",
                    "LifeCycleConfig": {
                        "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                        "OnCreate": "filename",
                    },
                    "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                    "ThreadsPerCore": 1,
                },
            ],
            VpcConfig={
                "SecurityGroupIds": [
                    "sg-12345678901234567",
                ],
                "Subnets": [
                    "subnet-12345678901234567",
                ],
            },
            Tags=[
                {"Key": "testkey", "Value": "testvalue"},
            ],
        )
    assert (
        str(e.value)
        == "An error occurred (ResourceInUse) when calling the CreateCluster operation: Resource Already Exists: Cluster with name testcluster already exists. Choose a different name."
    )


@mock_aws
def test_create_cluster_bad_source_s3_uri():
    client = boto3.client("sagemaker", region_name="us-east-1")
    with pytest.raises(ClientError) as e:
        client.create_cluster(
            ClusterName="testcluster",
            InstanceGroups=[
                {
                    "InstanceCount": 10,
                    "InstanceGroupName": "testgroup",
                    "InstanceType": "ml.p4d.24xlarge",
                    "LifeCycleConfig": {
                        "SourceS3Uri": "s3://lifecycleconfig",
                        "OnCreate": "filename",
                    },
                    "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                    "ThreadsPerCore": 2,
                },
                {
                    "InstanceCount": 15,
                    "InstanceGroupName": "testgroup2",
                    "InstanceType": "ml.g5.8xlarge",
                    "LifeCycleConfig": {
                        "SourceS3Uri": "s3://sagemaker-lifecycleconfig2",
                        "OnCreate": "filename",
                    },
                    "ExecutionRole": "arn:aws:iam::123456789012:role/service-role/AmazonSageMaker-TestExecutionRole",
                    "ThreadsPerCore": 1,
                },
            ],
            VpcConfig={
                "SecurityGroupIds": [
                    "sg-12345678901234567",
                ],
                "Subnets": [
                    "subnet-12345678901234567",
                ],
            },
            Tags=[
                {"Key": "testkey", "Value": "testvalue"},
            ],
        )
    assert (
        str(e.value)
        == "An error occurred (ValidationException) when calling the CreateCluster operation: Validation Error: SourceS3Uri s3://lifecycleconfig does not start with 's3://sagemaker'."
    )



"""
This file should only contain constants used for the EKS tests.
"""

import re
from enum import Enum

from boto3 import Session

from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

DEFAULT_ENCODING = "utf-8"
DEFAULT_HTTP_HEADERS = {"Content-type": "application/json"}
DEFAULT_NAMESPACE = "namespace_1"
FROZEN_TIME = "2013-11-27T01:42:00Z"
MAX_FARGATE_LABELS = 5
PARTITIONS = Session().get_available_partitions()
REGION = "us-east-1"
SERVICE = "eks"
SUBNET_IDS = ["subnet-12345ab", "subnet-67890cd"]


AMI_TYPE_KEY = "amiType"
AMI_TYPE_VALUE = "AL2_x86_64"

CLIENT_REQUEST_TOKEN_KEY = "clientRequestToken"
CLIENT_REQUEST_TOKEN_VALUE = "test_request_token"

DISK_SIZE_KEY = "diskSize"
DISK_SIZE_VALUE = 30

ENCRYPTION_CONFIG_KEY = "encryptionConfig"
ENCRYPTION_CONFIG_VALUE = [
    {"resources": ["secrets"], "provider": {"keyArn": "arn:of:the:key"}}
]

INSTANCE_TYPES_KEY = "instanceTypes"
INSTANCE_TYPES_VALUE = ["t3.medium"]

KUBERNETES_NETWORK_CONFIG_KEY = "kubernetesNetworkConfig"
KUBERNETES_NETWORK_CONFIG_VALUE = {"serviceIpv4Cidr": "172.20.0.0/16"}

LABELS_KEY = "labels"
LABELS_VALUE = {"purpose": "example"}

LAUNCH_TEMPLATE_KEY = "launchTemplate"
LAUNCH_TEMPLATE_VALUE = {"name": "myTemplate", "version": "2", "id": "123456"}

LOGGING_KEY = "logging"
LOGGING_VALUE = {"clusterLogging": [{"types": ["api"], "enabled": True}]}

NODEROLE_ARN_KEY = "nodeRole"
NODEROLE_ARN_VALUE = "arn:aws:iam::" + str(ACCOUNT_ID) + ":role/role_name"

POD_EXECUTION_ROLE_ARN_KEY = "podExecutionRoleArn"
POD_EXECUTION_ROLE_ARN_VALUE = "arn:aws:iam::" + str(ACCOUNT_ID) + ":role/role_name"

REMOTE_ACCESS_KEY = "remoteAccess"
REMOTE_ACCESS_VALUE = {"ec2SshKey": "eksKeypair"}

RESOURCES_VPC_CONFIG_KEY = "resourcesVpcConfig"
RESOURCES_VPC_CONFIG_VALUE = {
    "subnetIds": SUBNET_IDS,
    "endpointPublicAccess": True,
    "endpointPrivateAccess": False,
}

ROLE_ARN_KEY = "roleArn"
ROLE_ARN_VALUE = "arn:aws:iam::" + str(ACCOUNT_ID) + ":role/role_name"

SCALING_CONFIG_KEY = "scalingConfig"
SCALING_CONFIG_VALUE = {"minSize": 2, "maxSize": 3, "desiredSize": 2}

SELECTORS_KEY = "selectors"
SELECTORS_VALUE = [{"namespace": "profile-namespace"}]

STATUS_KEY = "status"
STATUS_VALUE = "ACTIVE"

SUBNETS_KEY = "subnets"
SUBNETS_VALUE = SUBNET_IDS

TAGS_KEY = "tags"
TAGS_VALUE = {"hello": "world"}

VERSION_KEY = "version"
VERSION_VALUE = "1"

AMI_TYPE = (AMI_TYPE_KEY, AMI_TYPE_VALUE)
CLIENT_REQUEST_TOKEN = (CLIENT_REQUEST_TOKEN_KEY, CLIENT_REQUEST_TOKEN_VALUE)
DISK_SIZE = (DISK_SIZE_KEY, DISK_SIZE_VALUE)
ENCRYPTION_CONFIG = (ENCRYPTION_CONFIG_KEY, ENCRYPTION_CONFIG_VALUE)
INSTANCE_TYPES = (INSTANCE_TYPES_KEY, INSTANCE_TYPES_VALUE)
KUBERNETES_NETWORK_CONFIG = (
    KUBERNETES_NETWORK_CONFIG_KEY,
    KUBERNETES_NETWORK_CONFIG_VALUE,
)
LABELS = (LABELS_KEY, LABELS_VALUE)
LAUNCH_TEMPLATE = (LAUNCH_TEMPLATE_KEY, LAUNCH_TEMPLATE_VALUE)
LOGGING = (LOGGING_KEY, LOGGING_VALUE)
NODEROLE_ARN = (NODEROLE_ARN_KEY, NODEROLE_ARN_VALUE)
POD_EXECUTION_ROLE_ARN = (POD_EXECUTION_ROLE_ARN_KEY, POD_EXECUTION_ROLE_ARN_VALUE)
REMOTE_ACCESS = (REMOTE_ACCESS_KEY, REMOTE_ACCESS_VALUE)
RESOURCES_VPC_CONFIG = (RESOURCES_VPC_CONFIG_KEY, RESOURCES_VPC_CONFIG_VALUE)
ROLE_ARN = (ROLE_ARN_KEY, ROLE_ARN_VALUE)
SCALING_CONFIG = (SCALING_CONFIG_KEY, SCALING_CONFIG_VALUE)
SELECTORS = (SELECTORS_KEY, SELECTORS_VALUE)
STATUS = (STATUS_KEY, STATUS_VALUE)
SUBNETS = (SUBNETS_KEY, SUBNETS_VALUE)
TAGS = (TAGS_KEY, TAGS_VALUE)
VERSION = (VERSION_KEY, VERSION_VALUE)


class ResponseAttributes:
    CLUSTER = "cluster"
    CLUSTERS = "clusters"
    FARGATE_PROFILE_NAMES = "fargateProfileNames"
    FARGATE_PROFILE = "fargateProfile"
    MESSAGE = "message"
    NEXT_TOKEN = "nextToken"
    NODEGROUP = "nodegroup"
    NODEGROUPS = "nodegroups"


class ErrorAttributes:
    CODE = "Code"
    ERROR = "Error"
    MESSAGE = "Message"


class ClusterInputs:
    REQUIRED = [ROLE_ARN, RESOURCES_VPC_CONFIG]
    OPTIONAL = [
        CLIENT_REQUEST_TOKEN,
        ENCRYPTION_CONFIG,
        LOGGING,
        KUBERNETES_NETWORK_CONFIG,
        TAGS,
        VERSION,
    ]


class FargateProfileInputs:
    REQUIRED = [POD_EXECUTION_ROLE_ARN, SELECTORS]
    OPTIONAL = [SUBNETS, TAGS]


class NodegroupInputs:
    REQUIRED = [NODEROLE_ARN, SUBNETS]
    OPTIONAL = [
        AMI_TYPE,
        DISK_SIZE,
        INSTANCE_TYPES,
        LABELS,
        REMOTE_ACCESS,
        SCALING_CONFIG,
        TAGS,
    ]


class PossibleTestResults(Enum):
    SUCCESS = "SUCCESS"
    FAILURE = "FAILURE"


class AddonAttributes:
    ADDON_NAME = "addonName"


class ClusterAttributes:
    ARN = "arn"
    CLUSTER_NAME = "clusterName"
    CREATED_AT = "createdAt"
    ENDPOINT = "endpoint"
    IDENTITY = "identity"
    ISSUER = "issuer"
    NAME = "name"
    OIDC = "oidc"
    ENCRYPTION_CONFIG = "encryptionConfig"


class FargateProfileAttributes:
    ARN = "fargateProfileArn"
    CREATED_AT = "createdAt"
    FARGATE_PROFILE_NAME = "fargateProfileName"
    LABELS = "labels"
    NAMESPACE = "namespace"
    SELECTORS = "selectors"


class NodegroupAttributes:
    ARN = "nodegroupArn"
    AUTOSCALING_GROUPS = "autoScalingGroups"
    CREATED_AT = "createdAt"
    MODIFIED_AT = "modifiedAt"
    NAME = "name"
    NODEGROUP_NAME = "nodegroupName"
    REMOTE_ACCESS_SG = "remoteAccessSecurityGroup"
    RESOURCES = "resources"


class BatchCountSize:
    SINGLE = 1
    SMALL = 10
    MEDIUM = 20
    LARGE = 200


class PageCount:
    SMALL = 3
    LARGE = 10


FARGATE_PROFILE_UUID_PATTERN = "(?P<fargate_uuid>[-0-9a-z]{8}-[-0-9a-z]{4}-[-0-9a-z]{4}-[-0-9a-z]{4}-[-0-9a-z]{12})"
NODEGROUP_UUID_PATTERN = "(?P<nodegroup_uuid>[-0-9a-z]{8}-[-0-9a-z]{4}-[-0-9a-z]{4}-[-0-9a-z]{4}-[-0-9a-z]{12})"


class RegExTemplates:
    CLUSTER_ARN = re.compile(
        "arn:"
        + "(?P<partition>.+):"
        + "eks:"
        + "(?P<region>[-0-9a-zA-Z]+):"
        + "(?P<account_id>[0-9]{12}):"
        + "cluster/"
        + "(?P<cluster_name>.+)"
    )
    FARGATE_PROFILE_ARN = re.compile(
        "arn:"
        + "(?P<partition>.+):"
        + "eks:"
        + "(?P<region>[-0-9a-zA-Z]+):"
        + "(?P<account_id>[0-9]{12}):"
        + "fargateprofile/"
        + "(?P<cluster_name>.+)/"
        + "(?P<fargate_name>.+)/"
        + FARGATE_PROFILE_UUID_PATTERN
    )
    ISO8601_FORMAT = re.compile(
        r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"
    )
    NODEGROUP_ARN = re.compile(
        "arn:"
        + "(?P<partition>.+):"
        + "eks:"
        + "(?P<region>[-0-9a-zA-Z]+):"
        + "(?P<account_id>[0-9]{12}):"
        + "nodegroup/"
        + "(?P<cluster_name>.+)/"
        + "(?P<nodegroup_name>.+)/"
        + NODEGROUP_UUID_PATTERN
    )
    NODEGROUP_ASG_NAME_PATTERN = re.compile("eks-" + NODEGROUP_UUID_PATTERN)
    NODEGROUP_SECURITY_GROUP_NAME_PATTERN = re.compile("sg-" + "([-0-9a-z]{17})")


class Endpoints:
    CREATE_CLUSTER = "/clusters"
    CREATE_NODEGROUP = "/clusters/{clusterName}/node-groups"
    DESCRIBE_CLUSTER = "/clusters/{clusterName}"
    DESCRIBE_NODEGROUP = "/clusters/{clusterName}/node-groups/{nodegroupName}"
    DELETE_CLUSTER = "/clusters/{clusterName}"
    DELETE_NODEGROUP = "/clusters/{clusterName}/node-groups/{nodegroupName}"
    LIST_CLUSTERS = "/clusters?maxResults={maxResults}&nextToken={nextToken}"
    LIST_NODEGROUPS = "/clusters/{clusterName}/node-groups?maxResults={maxResults}&nextToken={nextToken}"


class StatusCodes:
    OK = 200


class HttpHeaders:
    ErrorType = "x-amzn-ErrorType"



from __future__ import annotations

from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.assign.assign_decl import AssignDecl
from moto.stepfunctions.parser.asl.component.common.catch.catcher_outcome import (
    CatcherOutcomeCaught,
    CatcherOutcomeNotCaught,
)
from moto.stepfunctions.parser.asl.component.common.catch.catcher_props import (
    CatcherProps,
)
from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.error_name.error_equals_decl import (
    ErrorEqualsDecl,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.common.flow.next import Next
from moto.stepfunctions.parser.asl.component.common.outputdecl import Output
from moto.stepfunctions.parser.asl.component.common.path.result_path import ResultPath
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class CatcherOutput(dict):
    def __init__(self, error: str, cause: str):
        super().__init__()
        self["Error"] = error
        self["Cause"] = cause


class CatcherDecl(EvalComponent):
    DEFAULT_RESULT_PATH: ResultPath = ResultPath(result_path_src="$")

    error_equals: Final[ErrorEqualsDecl]
    next_decl: Next
    result_path: Optional[ResultPath]
    assign: Optional[AssignDecl]
    output: Optional[Output]
    comment: Final[Optional[Comment]]

    def __init__(
        self,
        error_equals: ErrorEqualsDecl,
        next_decl: Next,
        result_path: Optional[ResultPath],
        assign: Optional[AssignDecl],
        output: Optional[Output],
        comment: Optional[Comment],
    ):
        self.error_equals = error_equals
        self.next_decl = next_decl
        self.result_path = result_path
        self.assign = assign
        self.output = output
        self.comment = comment

    @classmethod
    def from_catcher_props(cls, props: CatcherProps) -> CatcherDecl:
        return cls(
            error_equals=props.get(
                typ=ErrorEqualsDecl,
                raise_on_missing=ValueError(
                    f"Missing ErrorEquals declaration for Catcher declaration, in props '{props}'."
                ),
            ),
            next_decl=props.get(
                typ=Next,
                raise_on_missing=ValueError(
                    f"Missing Next declaration for Catcher declaration, in props '{props}'."
                ),
            ),
            result_path=props.get(typ=ResultPath),
            assign=props.get(typ=AssignDecl),
            output=props.get(typ=Output),
            comment=props.get(typ=Comment),
        )

    def _eval_body(self, env: Environment) -> None:
        failure_event: FailureEvent = env.stack.pop()

        env.stack.append(failure_event.error_name)
        self.error_equals.eval(env)

        equals: bool = env.stack.pop()
        if equals:
            # Input for the catch block is the error output.
            env.stack.append(env.states.get_error_output())

            if self.assign:
                self.assign.eval(env=env)

            if self.result_path:
                self.result_path.eval(env)

            # Prepare the state output: successful catch states override the states' output procedure.
            if self.output:
                self.output.eval(env=env)
            else:
                output_value = env.stack.pop()
                env.states.reset(output_value)

            # Append successful output to notify the outcome upstream.
            env.next_state_name = self.next_decl.name
            env.stack.append(CatcherOutcomeCaught())
        else:
            env.stack.append(failure_event)
            env.stack.append(CatcherOutcomeNotCaught())



"""Exceptions raised by the timestreaminfluxdb service."""

from moto.core.exceptions import JsonRESTError


class TimestreamInfluxDBException(JsonRESTError):
    pass


class ValidationException(TimestreamInfluxDBException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class ConflictException(TimestreamInfluxDBException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ConflictException", message)


class ResourceNotFoundException(TimestreamInfluxDBException):
    code = 400

    def __init__(self, message: str):
        super().__init__("ResourceNotFoundException", message)



import boto3
import botocore.exceptions
import pytest

from moto import mock_aws


@mock_aws
def test_create_api_mapping():
    client = boto3.client("apigatewayv2", region_name="us-east-1")
    api = client.create_api(Name="test-api", ProtocolType="HTTP")
    dev_domain = client.create_domain_name(DomainName="dev.service.io")
    expected_keys = ["ApiId", "ApiMappingId", "ApiMappingKey", "Stage"]

    post_resp = client.create_api_mapping(
        DomainName=dev_domain["DomainName"],
        ApiMappingKey="v1/api",
        Stage="$default",
        ApiId=api["ApiId"],
    )

    get_resp = client.get_api_mapping(
        DomainName="dev.service.io", ApiMappingId=post_resp["ApiMappingId"]
    )

    # check post response has all expected keys
    for key in expected_keys:
        assert key in post_resp

    # check get response has all expected keys
    for key in expected_keys:
        assert key in get_resp

    # check that values are equal for post and get of same resource
    for key in expected_keys:
        assert post_resp.get(key) == get_resp.get(key)

    # ensure known values are set correct in post response
    assert post_resp.get("ApiId") is not None
    assert post_resp.get("ApiMappingId") is not None
    assert post_resp.get("ApiMappingKey") == "v1/api"
    assert post_resp.get("Stage") == "$default"


@mock_aws
def test_create_api_mapping_missing_api():
    client = boto3.client("apigatewayv2", region_name="us-east-1")
    dev_domain = client.create_domain_name(DomainName="dev.service.io")
    with pytest.raises(botocore.exceptions.ClientError) as exc:
        client.create_api_mapping(
            DomainName=dev_domain["DomainName"],
            Stage="$default",
            ApiId="123dne",
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert (
        err["Message"]
        == "Invalid API identifier specified The resource specified in the request was not found."
    )


@mock_aws
def test_create_api_mapping_missing_domain():
    client = boto3.client("apigatewayv2", region_name="us-east-1")
    api = client.create_api(Name="test-api", ProtocolType="HTTP")

    with pytest.raises(botocore.exceptions.ClientError) as exc:
        client.create_api_mapping(
            DomainName="domain.dne",
            Stage="$default",
            ApiId=api["ApiId"],
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert (
        err["Message"]
        == "The domain name resource specified in the request was not found."
    )


@mock_aws
def test_create_api_mapping_bad_mapping_keys():
    client = boto3.client("apigatewayv2", region_name="us-east-1")
    api = client.create_api(Name="test-api", ProtocolType="HTTP")
    dev_domain = client.create_domain_name(DomainName="dev.service.io")
    bad_keys = {
        "/api": "API mapping key should not start with a '/' or have consecutive '/'s.",
        "v1//api": "API mapping key should not start with a '/' or have consecutive '/'s.",
        "api/": "API mapping key should not end with a '/'.",
    }

    for bad_key, message in bad_keys.items():
        with pytest.raises(botocore.exceptions.ClientError) as exc:
            client.create_api_mapping(
                DomainName=dev_domain["DomainName"],
                ApiMappingKey=bad_key,
                Stage="$default",
                ApiId=api["ApiId"],
            )

        err = exc.value.response["Error"]
        assert err["Code"] == "BadRequestException"
        assert err["Message"] == message


@mock_aws
def test_get_api_mappings():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api = client.create_api(Name="test-api", ProtocolType="HTTP")
    included_domain = client.create_domain_name(DomainName="dev.service.io")
    excluded_domain = client.create_domain_name(DomainName="hr.service.io")

    v1_mapping = client.create_api_mapping(
        DomainName=included_domain["DomainName"],
        ApiMappingKey="v1/api",
        Stage="$default",
        ApiId=api["ApiId"],
    )

    v2_mapping = client.create_api_mapping(
        DomainName=included_domain["DomainName"],
        ApiMappingKey="v2/api",
        Stage="$default",
        ApiId=api["ApiId"],
    )

    hr_mapping = client.create_api_mapping(
        DomainName=excluded_domain["DomainName"],
        ApiMappingKey="hr/api",
        Stage="$default",
        ApiId=api["ApiId"],
    )

    # sanity check responses
    assert v1_mapping["ApiMappingKey"] == "v1/api"
    assert v2_mapping["ApiMappingKey"] == "v2/api"
    assert hr_mapping["ApiMappingKey"] == "hr/api"

    # make comparable
    del v1_mapping["ResponseMetadata"]
    del v2_mapping["ResponseMetadata"]
    del hr_mapping["ResponseMetadata"]

    get_resp = client.get_api_mappings(DomainName=included_domain["DomainName"])
    assert "Items" in get_resp
    assert v1_mapping in get_resp.get("Items")
    assert v2_mapping in get_resp.get("Items")
    assert hr_mapping not in get_resp.get("Items")


@mock_aws
def test_delete_api_mapping():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api = client.create_api(Name="test-api", ProtocolType="HTTP")
    api_domain = client.create_domain_name(DomainName="dev.service.io")

    v1_mapping = client.create_api_mapping(
        DomainName=api_domain["DomainName"],
        ApiMappingKey="v1/api",
        Stage="$default",
        ApiId=api["ApiId"],
    )

    del v1_mapping["ResponseMetadata"]

    get_resp = client.get_api_mappings(DomainName=api_domain["DomainName"])
    assert "Items" in get_resp
    assert v1_mapping in get_resp.get("Items")

    client.delete_api_mapping(
        DomainName=api_domain["DomainName"], ApiMappingId=v1_mapping["ApiMappingId"]
    )

    get_resp = client.get_api_mappings(DomainName=api_domain["DomainName"])
    assert "Items" in get_resp
    assert v1_mapping not in get_resp.get("Items")


@mock_aws
def test_delete_api_mapping_dne():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    client.create_api(Name="test-api", ProtocolType="HTTP")
    api_domain = client.create_domain_name(DomainName="dev.service.io")

    with pytest.raises(botocore.exceptions.ClientError) as exc:
        client.delete_api_mapping(
            DomainName=api_domain["DomainName"], ApiMappingId="123dne"
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "NotFoundException"
    assert (
        err["Message"]
        == "The api mapping resource specified in the request was not found."
    )



import boto3

from moto import mock_aws


@mock_aws
def test_create_without_tags():
    client = boto3.client("es", region_name="us-east-1")
    arn = client.create_elasticsearch_domain(DomainName="testdn")["DomainStatus"]["ARN"]

    assert client.list_tags(ARN=arn)["TagList"] == []


@mock_aws
def test_create_with_tags():
    client = boto3.client("es", region_name="us-east-1")
    domain = client.create_elasticsearch_domain(
        DomainName="testdn", TagList=[{"Key": "k1", "Value": "v1"}]
    )
    arn = domain["DomainStatus"]["ARN"]

    assert client.list_tags(ARN=arn)["TagList"] == [{"Key": "k1", "Value": "v1"}]


@mock_aws
def test_add_tags():
    client = boto3.client("es", region_name="us-east-1")
    domain = client.create_elasticsearch_domain(DomainName="testdn")

    arn = domain["DomainStatus"]["ARN"]

    client.add_tags(
        ARN=arn, TagList=[{"Key": "k1", "Value": "v1"}, {"Key": "k2", "Value": "v2"}]
    )

    assert client.list_tags(ARN=arn)["TagList"] == [
        {"Key": "k1", "Value": "v1"},
        {"Key": "k2", "Value": "v2"},
    ]


@mock_aws
def test_update_tag():
    client = boto3.client("es", region_name="us-east-1")
    domain = client.create_elasticsearch_domain(
        DomainName="testdn", TagList=[{"Key": "k1", "Value": "v1"}]
    )

    arn = domain["DomainStatus"]["ARN"]
    assert client.list_tags(ARN=arn)["TagList"] == [{"Key": "k1", "Value": "v1"}]

    # add the same key again with a different value
    client.add_tags(ARN=arn, TagList=[{"Key": "k1", "Value": "v2"}])

    assert client.list_tags(ARN=arn)["TagList"] == [
        {"Key": "k1", "Value": "v2"},
    ]


@mock_aws
def test_remove_tags():
    client = boto3.client("es", region_name="us-east-1")
    domain = client.create_elasticsearch_domain(
        DomainName="testdn",
        TagList=[{"Key": "k1", "Value": "v1"}, {"Key": "k2", "Value": "v2"}],
    )
    arn = domain["DomainStatus"]["ARN"]

    assert client.list_tags(ARN=arn)["TagList"] == [
        {"Key": "k1", "Value": "v1"},
        {"Key": "k2", "Value": "v2"},
    ]

    client.remove_tags(ARN=arn, TagKeys=["k1"])
    assert client.list_tags(ARN=arn)["TagList"] == [{"Key": "k2", "Value": "v2"}]

    client.remove_tags(ARN=arn, TagKeys=["k2"])
    assert client.list_tags(ARN=arn)["TagList"] == []



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_create_authorizer_minimum():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    resp = client.create_authorizer(
        ApiId=api_id,
        AuthorizerType="REQUEST",
        IdentitySource=[],
        Name="auth1",
        AuthorizerPayloadFormatVersion="2.0",
    )

    assert "AuthorizerId" in resp
    assert resp["AuthorizerType"] == "REQUEST"
    assert resp["Name"] == "auth1"


@mock_aws
def test_create_authorizer():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    resp = client.create_authorizer(
        ApiId=api_id,
        AuthorizerCredentialsArn="auth:creds:arn",
        AuthorizerPayloadFormatVersion="2.0",
        AuthorizerResultTtlInSeconds=3,
        AuthorizerType="REQUEST",
        AuthorizerUri="auth_uri",
        EnableSimpleResponses=True,
        IdentitySource=["$request.header.Authorization"],
        IdentityValidationExpression="ive",
        JwtConfiguration={"Audience": ["a1"], "Issuer": "moto.com"},
        Name="auth1",
    )

    assert "AuthorizerId" in resp
    assert resp["AuthorizerCredentialsArn"] == "auth:creds:arn"
    assert resp["AuthorizerPayloadFormatVersion"] == "2.0"
    assert resp["AuthorizerResultTtlInSeconds"] == 3
    assert resp["AuthorizerType"] == "REQUEST"
    assert resp["AuthorizerUri"] == "auth_uri"
    assert resp["EnableSimpleResponses"] is True
    assert resp["IdentitySource"] == ["$request.header.Authorization"]
    assert resp["IdentityValidationExpression"] == "ive"
    assert resp["JwtConfiguration"] == {"Audience": ["a1"], "Issuer": "moto.com"}
    assert resp["Name"] == "auth1"
    assert resp["AuthorizerPayloadFormatVersion"] == "2.0"


@mock_aws
def test_create_authorizer_without_payloadformatversion():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    with pytest.raises(ClientError) as exc:
        client.create_authorizer(
            ApiId=api_id,
            AuthorizerType="REQUEST",
            AuthorizerUri="auth_uri",
            IdentitySource=[""],
            Name="auth1",
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "BadRequestException"
    assert (
        err["Message"]
        == "AuthorizerPayloadFormatVersion is a required parameter for REQUEST authorizer"
    )


@mock_aws
def test_get_authorizer():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    authorizer_id = client.create_authorizer(
        ApiId=api_id,
        AuthorizerType="REQUEST",
        IdentitySource=[],
        Name="auth1",
        AuthorizerPayloadFormatVersion="2.0",
    )["AuthorizerId"]

    resp = client.get_authorizer(ApiId=api_id, AuthorizerId=authorizer_id)

    assert "AuthorizerId" in resp
    assert resp["AuthorizerType"] == "REQUEST"
    assert resp["Name"] == "auth1"
    assert resp["AuthorizerPayloadFormatVersion"] == "2.0"


@mock_aws
def test_delete_authorizer():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="WEBSOCKET")["ApiId"]

    authorizer_id = client.create_authorizer(
        ApiId=api_id, AuthorizerType="REQUEST", IdentitySource=[], Name="auth1"
    )["AuthorizerId"]

    client.delete_authorizer(ApiId=api_id, AuthorizerId=authorizer_id)

    with pytest.raises(ClientError) as exc:
        client.get_authorizer(ApiId=api_id, AuthorizerId="unknown")

    err = exc.value.response["Error"]
    assert err["Code"] == "NotFoundException"


@mock_aws
def test_get_authorizer_unknown():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    with pytest.raises(ClientError) as exc:
        client.get_authorizer(ApiId=api_id, AuthorizerId="unknown")

    err = exc.value.response["Error"]
    assert err["Code"] == "NotFoundException"


@mock_aws
def test_update_authorizer_single():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    auth_id = client.create_authorizer(
        ApiId=api_id,
        AuthorizerCredentialsArn="auth:creds:arn",
        AuthorizerPayloadFormatVersion="2.0",
        AuthorizerResultTtlInSeconds=3,
        AuthorizerType="REQUEST",
        AuthorizerUri="auth_uri",
        EnableSimpleResponses=True,
        IdentitySource=["$request.header.Authorization"],
        IdentityValidationExpression="ive",
        JwtConfiguration={"Audience": ["a1"], "Issuer": "moto.com"},
        Name="auth1",
    )["AuthorizerId"]

    resp = client.update_authorizer(ApiId=api_id, AuthorizerId=auth_id, Name="auth2")

    assert "AuthorizerId" in resp
    assert resp["AuthorizerCredentialsArn"] == "auth:creds:arn"
    assert resp["AuthorizerPayloadFormatVersion"] == "2.0"
    assert resp["AuthorizerResultTtlInSeconds"] == 3
    assert resp["AuthorizerType"] == "REQUEST"
    assert resp["AuthorizerUri"] == "auth_uri"
    assert resp["EnableSimpleResponses"] is True
    assert resp["IdentitySource"] == ["$request.header.Authorization"]
    assert resp["IdentityValidationExpression"] == "ive"
    assert resp["JwtConfiguration"] == {"Audience": ["a1"], "Issuer": "moto.com"}
    assert resp["Name"] == "auth2"


@mock_aws
def test_update_authorizer_all_attributes():
    client = boto3.client("apigatewayv2", region_name="eu-west-1")
    api_id = client.create_api(Name="test-api", ProtocolType="HTTP")["ApiId"]

    auth_id = client.create_authorizer(
        ApiId=api_id,
        AuthorizerType="REQUEST",
        IdentitySource=[],
        Name="auth1",
        AuthorizerPayloadFormatVersion="2.0",
    )["AuthorizerId"]

    auth_id = client.update_authorizer(
        ApiId=api_id,
        AuthorizerId=auth_id,
        AuthorizerCredentialsArn="",
        AuthorizerPayloadFormatVersion="3.0",
        AuthorizerResultTtlInSeconds=5,
        AuthorizerType="REQUEST",
        AuthorizerUri="auth_uri",
        EnableSimpleResponses=False,
        IdentitySource=["$request.header.Authentication"],
        IdentityValidationExpression="ive2",
        JwtConfiguration={"Audience": ["a2"], "Issuer": "moto.com"},
        Name="auth1",
    )["AuthorizerId"]

    resp = client.update_authorizer(ApiId=api_id, AuthorizerId=auth_id, Name="auth2")

    assert "AuthorizerId" in resp
    assert resp["AuthorizerCredentialsArn"] == ""
    assert resp["AuthorizerPayloadFormatVersion"] == "3.0"
    assert resp["AuthorizerResultTtlInSeconds"] == 5
    assert resp["AuthorizerType"] == "REQUEST"
    assert resp["AuthorizerUri"] == "auth_uri"
    assert resp["EnableSimpleResponses"] is False
    assert resp["IdentitySource"] == ["$request.header.Authentication"]
    assert resp["IdentityValidationExpression"] == "ive2"
    assert resp["JwtConfiguration"] == {"Audience": ["a2"], "Issuer": "moto.com"}
    assert resp["Name"] == "auth2"



from functools import wraps
from uuid import uuid4

import boto3

from moto import mock_aws
from tests import allow_aws_request


def iam_aws_verified(create_user: bool = False):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.
    """

    def inner(func):
        def create_user_and_invoke_test():
            client = boto3.client("iam", "us-east-1")
            user_name = f"testuser_{str(uuid4())[0:6]}"
            try:
                if create_user:
                    client.create_user(UserName=user_name)
                return func(user_name=user_name)
            finally:
                if create_user:
                    certificates = client.list_signing_certificates(UserName=user_name)[
                        "Certificates"
                    ]

                    for cert in certificates:
                        client.delete_signing_certificate(
                            UserName=user_name, CertificateId=cert["CertificateId"]
                        )
                    client.delete_user(UserName=user_name)

        @wraps(func)
        def pagination_wrapper():
            if allow_aws_request():
                return create_user_and_invoke_test()
            else:
                with mock_aws():
                    return create_user_and_invoke_test()

        return pagination_wrapper

    return inner



import unittest

import boto3

from moto import mock_aws


@mock_aws
class TestDynamoDBTagging(unittest.TestCase):
    def setUp(self) -> None:
        self.dynamodb = boto3.client("dynamodb", region_name="us-west-2")
        self.rtapi = boto3.client("resourcegroupstaggingapi", region_name="us-west-2")
        self.resources_tagged = []
        self.resources_untagged = []
        for i in range(3):
            table = self.dynamodb.create_table(
                AttributeDefinitions=[{"AttributeName": "col", "AttributeType": "S"}],
                TableName=f"table-{i}",
                KeySchema=[{"AttributeName": "col", "KeyType": "HASH"}],
                BillingMode="PAY_PER_REQUEST",
            ).get("TableDescription")
            self.dynamodb.tag_resource(
                ResourceArn=table["TableArn"],
                Tags=[{"Key": "test", "Value": f"value-{i}"}] if i else [],
            )
            group = self.resources_tagged if i else self.resources_untagged
            group.append(table["TableArn"])

    def test_get_resources_dynamodb(self):
        def assert_response(response, expected_count, resource_type=None):
            results = response.get("ResourceTagMappingList", [])
            assert len(results) == expected_count
            for item in results:
                arn = item["ResourceARN"]
                assert arn in self.resources_tagged
                assert arn not in self.resources_untagged
                if resource_type:
                    assert f":{resource_type}:" in arn

        resp = self.rtapi.get_resources(ResourceTypeFilters=["dynamodb"])
        assert_response(resp, 2)
        resp = self.rtapi.get_resources(ResourceTypeFilters=["dynamodb:table"])
        assert_response(resp, 2)
        resp = self.rtapi.get_resources(
            TagFilters=[{"Key": "test", "Values": ["value-1"]}]
        )
        assert_response(resp, 1)

    def test_tag_resources_dynamodb(self):
        # WHEN
        # we tag resources
        self.rtapi.tag_resources(
            ResourceARNList=self.resources_tagged,
            Tags={"key1": "value1", "key2": "value2"},
        )

        # THEN
        # we can retrieve the tags using the DynamoDB API
        def get_tags(arn):
            return self.dynamodb.list_tags_of_resource(ResourceArn=arn)["Tags"]

        for arn in self.resources_tagged:
            assert {"Key": "key1", "Value": "value1"} in get_tags(arn)
            assert {"Key": "key2", "Value": "value2"} in get_tags(arn)
        for arn in self.resources_untagged:
            assert get_tags(arn) == []
